/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on Jan 13, 2005
 */
package com.sunlabs.archsim.gasp;

import com.sunlabs.archsim.base.hierarchy.SimpleEntity;
import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.globals.GlobalDefaults;
import com.sunlabs.archsim.globals.Globals;
// import com.sunlabs.archsim.framework.Function;

/**
 * @author ib27688
 *
 * Generic model for the state. 
 * @see SML 2005-379.
 */


import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

public class State extends SimpleEntity {
	
	/** global settings */
	protected Globals globals= Globals.getInstance();
		
	/** Executing this command makes the state active */
	Command setActiveCommand;
	/** Executing this command makes the state inactive */
	Command setInactiveCommand;

	/** Keep track whether the state is active or inactive */
	protected boolean active;
	
	/** inital activity */
	protected boolean initiallyActive;
	
	/** initial data - if active */
	protected Object initialData;
	
	/** The data associated with the state.
	 *  This data is only "valid" when the state is active.
	 */
	protected Object data;
	
	/** does this state need to journal its sctivity? */
	protected boolean journalState;
	
	/** The set of terminals attached to the state.
	 * The set is used as we allow no duplicates.
	 */
	protected Set<Terminal> terminals;
	
	/** create an initially inactove state */
	public State(String name) {
		super(name);
		terminals= new HashSet<Terminal>();
		active= false;
		
		initiallyActive= false;
		initialData= null;
		
		data= null;
		setActiveCommand= new SetActiveCommand();
		setInactiveCommand= new SetInActiveCommand();
		
		Boolean jState= globals.booleanValue(GlobalDefaults.JOURNAL_STATE);
		journalState= (jState != null)
			? jState.booleanValue()
			: GlobalDefaults.JOURNAL_STATE_DEF;		
	} // State constructor
	
	/** Create an initailly active state. 
	 * Note that the state will become active when the init
	 * method is called.
	 */
	public State(String name, Object data) {
		this(name);
		setInitiallyActive(data);
	}
	

	public State(String name, CompositeEntity g) {
		this(name);
		setParent(g);
	} // State constructor

/*	public State(String name, Group g, Object data) {
		this(name, g);
		setInitiallyActive(data);
	} // State constructor
*/	
	
	public String stateInfo() {
		
		String status;
		if (active) {
			status = "State " + getName() + ", id= " + getID() + " is active with data " 
			+ data;
		}
		else {
			status= "State " + getName() + ", id= " + getID() + " is inactive";
		}
		
		return status;
	} // stateInfo
	
	
	/** set state initially active */
	public void setInitiallyInactive() {
		initiallyActive= false;
	}
	
	/** set state initially inactive, provide initial data */
	public void setInitiallyActive(Object data) {
		initiallyActive= true;
		initialData= data;
	}
	
	
	public void attach(Terminal t) {
		// add the terminal to the list
		// TODO
		// ignore the boolean return - the consistency check
		// will ensure connections are done properly
		// we might rethink this with time		
		terminals.add(t);
		// let the terminal know it is attached to this state
		t.attachState(this);
	} // attachTerminal
			
	private class SetActiveCommand extends Command {
		
		/** when setting the state active, one must
		 *  provide the data that is "captured" by the state
		 */
		public void execute(Object d) {
			if (active) {
				director.fatalError("State " + getAlias() + 
						", id= " + getID() + 
						", made active without being inactive in between");
			}
			else {
				active= true;
				data= d;
				int logLevel;
				if (journalState) {
					// globals.print();
					Integer lLevel= globals.intValue(GlobalDefaults.LOG_LEVEL);
					logLevel= (lLevel != null)
						? lLevel.intValue()
						: GlobalDefaults.LOG_LEVEL_DEF;
					if (logLevel == GlobalDefaults.logDetails) {
						director.journal("State " + getAlias()
							+ ", id= " + getID() + 
							" is active with data " + d);
					}
					else if (logLevel == GlobalDefaults.logForPlayback) {
						director.journal(getJournalName(), GlobalDefaults.ACTIVE);
					}
				}
				// to all attached actions: the state is active
				sendNotifications();
			}
		} // execute
		
		/** force active,  intended for reset */
		public void forceActive(Object d) {
			boolean change= (active == false);
			active= true;
			data= d;
			int logLevel;
			if (journalState) {
				Integer lLevel= globals.intValue(GlobalDefaults.LOG_LEVEL);
				logLevel= (lLevel != null)
					? lLevel.intValue()
					: GlobalDefaults.LOG_LEVEL_DEF;
				if (logLevel == GlobalDefaults.logDetails) {
					director.journal("State " + getAlias()
						+ ", id= " + getID() + 
						" is INITIALIZED to active with data " + d);
				}
				else if (logLevel == GlobalDefaults.logForPlayback) {
					director.journal(getJournalName(), GlobalDefaults.ACTIVE);
				}
			}
			// to all attached actions: the state is active
			if (change) sendNotifications();			
		} // force active
		
	} // class SetActiveCommand
	
	private class SetInActiveCommand extends Command {
		
		public void execute(Object params) {
			if (!active) {
				director.fatalError("State " + getAlias() + 
						", id= " + getID() + 
						", made inactive without being active in between");
				director.stop();
				
			}
			else {
				int logLevel;
				active= false;
				if (journalState) {
					// globals.print();
					Integer lLevel= globals.intValue(GlobalDefaults.LOG_LEVEL);
					logLevel= (lLevel != null)
						? lLevel.intValue()
						: GlobalDefaults.LOG_LEVEL_DEF;				
					if (logLevel == GlobalDefaults.logDetails) {
						director.journal("State " + getAlias()
							+ ", id= " + getID() + 
							" is INactive");
					}
					else if (logLevel == GlobalDefaults.logForPlayback) {
						director.journal(getJournalName(), GlobalDefaults.INACTIVE);
					}
				}			
				// to all attached actions: the state is inactive
				sendNotifications();
			}
		} // execute
		
		/** force inactive - intended for reset */
		public void forceInactive() {
			int logLevel;
			boolean change= (active == true);
			active= false;
			if (journalState) {
				Integer lLevel= globals.intValue(GlobalDefaults.LOG_LEVEL);
				logLevel= (lLevel != null)
					? lLevel.intValue()
					: GlobalDefaults.LOG_LEVEL_DEF;				
				if (logLevel == GlobalDefaults.logDetails) {
					director.journal("State " + getAlias()
						+ ", id= " + getID() + 
						" is INITIALIZED to INactive");
				}
				else if (logLevel == GlobalDefaults.logForPlayback) {
					director.journal(getJournalName(), GlobalDefaults.INACTIVE);
				}
			}			
			// to all attached actions: the state is inactive
			if (change) sendNotifications();			
		} // forceInactive
		
	} // class SetInActiveCommand
	
	/** the state has changed, send updates through all
	 *  attached terminals. 
	 *  @param status
	 */
	private void sendNotifications() {
		Iterator terminalIterator= terminals.iterator();
		while (terminalIterator.hasNext()) {
			Terminal t= (Terminal) (terminalIterator.next());
			if (active) {
				// notify the attached terminal
				// the terminal will notify the action
				t.notifyStateIsActive();
			}
			else {
				// let the attached terminal the state is inactive
				// the terminal will notify the state
				t.notifyStateIsInactive();
			}
		}
	} // sendNotifications
	
	public void reset() {
		/** TODO fill in the blanks
		 * Are there any blanks anyway? 
		 */
	} // reset
	
	/**
	 * Check whethet the state is active
	 * @return true is the state is active, false otherwise
	 */
	public boolean isActive() {
		return active;
	} // isActive
	
	/** 
	 * Get the data that is captured by the state
	 * @return state data
	 */
	public Object getData() {
		return data;
	} // getData

	public void setActive(Object data) {
		setActiveCommand.execute(data);
	} // setActive
	
	public void setInactive() {
		setInactiveCommand.execute(null);
	} // setInactive
	
	/** turn journal for this particular state on/off */
	public void setJournal(boolean journal) {
		journalState= journal;
/*
 		int logLevel;
		Integer lLevel= globals.intValue(GlobalDefaults.LOG_LEVEL);
		logLevel= (lLevel != null)
			? lLevel.intValue()
			: GlobalDefaults.LOG_LEVEL_DEF;				


		// display the current state if journalling was just turned on		
		if (journalState) {
			
			if (active) {
				if (logLevel == GlobalDefaults.logDetails) {
					director.journal("State " + getAlias()
						+ ", id= " + getID() + 
						" is active with data " + data);
				}
				else if (logLevel == GlobalDefaults.logForPlayback) {
					director.journal(getJournalName(), GlobalDefaults.ACTIVE);
				}				
			}
			else {
				if (logLevel == GlobalDefaults.logDetails) {
					director.journal("State " + getAlias()
						+ ", id= " + getID() + 
						" is INactive");
				}
				else if (logLevel == GlobalDefaults.logForPlayback) {
					director.journal(getJournalName(), GlobalDefaults.INACTIVE);
				}	
			}
		}
*/
	} // setJournal
	
	public boolean isAttached(Terminal t) {
		return terminals.contains(t);
	} // isAttached
	
	/** Check for all attached terminals whether they are attached
	 * to this state.
	 * 
	 * @return true if all attachemnts are OK, false otherwise
	 */
	public boolean selfCheck() {
		boolean OK= true;
		Iterator ti= terminals.iterator();
		while (OK && ti.hasNext()) {
			Terminal t= (Terminal)(ti.next());
			OK= OK && t.isAttached(this);
		}
		return OK;
	} // selfCheck
	
	public void print() {
		director.info("State " + getAlias() + ", id= " + getID() 
				+ " is attached to terminals: ");
		Iterator ti= terminals.iterator();
		while (ti.hasNext()) {
			Terminal t= (Terminal)(ti.next());
			director.info("\t"+ t.getAlias() + ",id= " + t.getID());
		}
		if (isActive()) {
			director.info("\t" + "is active with data " + getData());
		}
		else {
			director.info("\t" + "is INactive");			
		}

	} // printConnections

	/** Entity initialization.
	 * Set state to what it should be initially. 
	 */
	public void init() {
		if (initiallyActive) {
			((SetActiveCommand)setActiveCommand).forceActive(initialData);
		}
		else {
			((SetInActiveCommand)setInactiveCommand).forceInactive();
		}
	}
	
} // class State
