/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.gasp;

import java.util.Iterator;

import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * An action with a trigger.
 * This action operates the same as the regular action, but it also requires a
 * enable trigger command to be executed before this action can fire. A disable trigger command
 * will prevent such an action from further firing, until the subsequent enable trigger.
 * The trigger, thus, acts as an on/off switch.
 * 
 * Initially, the "switch is off". That is, a TriggerAction will NOT fire until enable-trigger
 * command is executed.
 * 
 * Such an action is useful for constructing complex components, where multiple actions can
 * enable or disable each other.
 * 
 * @author ib27688
 *
 */

public abstract class TriggerAction extends Action {

	/** trigger flag */
	protected boolean enabled;

	/** enable trigger command */
	public Command enableTriggerCommand;
	
	/** disable trigger command */
	public Command disableTriggerCommand;
	
	public TriggerAction(String name) {
		super(name);
		
		// initially disabled
		enabled= false;
		// make the commands
		enableTriggerCommand= new EnableTriggerCommand();
		disableTriggerCommand= new DisableTriggerCommand();
	}
	
	/** 
	 * Process trigger parameters.
	 * Do nothing by default, but overrriiding this method provides
	 * an opportunity to set up action parameters for subsequent firing.
	 * 
	 * 
	 * @param params parameters to be processed
	 */
	protected void processEnableTrigger(Object params) {
		// do nothing by default
	}
	
	/** 
	 * Process trigger parameters.
	 * Do nothing by default, but overrriiding this method provides
	 * an opportunity to reset set up action parameters, or set them up
	 * for a subsequent enabling.
	 *
	 * @param params parameters to be processed
	 */	
	protected void processDisableTrigger(Object params) {
		// do nothing by default
	}
	
	/** this action can only fire when it is enabled */
	protected boolean canFire() {
		return enabled && super.canFire();
	}
	
	/** 
	 * Execution of this command enables firing.
	 * Note that, once firing is enabled, one should check if all inputs 
	 * are active, and schedule firing when this is the case.
	 *
	 */
	protected class EnableTriggerCommand extends Command {
		public void execute(Object params) {
			if (journalAction) {
				director.journal("Trigger action " + getAlias() 
						+ ", id= " + getID() 
						+ " was enabled");						
			}			
			// FIRST process trigger parameters - this order
			// allows for dealing with re-enabling
			processEnableTrigger(params);
			// now enable the action
			enabled= true;
			// check whether the acion can fire
			// this code is a copy of the code from StateActiveCommand
			// TODO refactoring opportunity
			if (canFire()) {
				// it is OK to fire
				// iterate though outputs and schedule firing
				Iterator ti= terminals.iterator();
				while (ti.hasNext()) {
					Terminal t= (Terminal) (ti.next());
					if (t instanceof OutputTerminal) {
						computeAndFire((OutputTerminal)t);
					}
					else if (t instanceof InputTerminal
							&& t.getRole() == Terminal.SELF_RESETTING_INPUT) {
						// deactivate all states attached to input terminals
						t.setStateInactive();
						if (journalAction) {
							director.journal("Action " + getAlias()
									+ ", id= " + getID()
									+ ", fired on terminal " 
									+ t.getAlias() + ", id= " + t.getID() 
									+ " to set state inactive");					
						}						
					} // else if input
				} // while hasNext
			} // if canFire	
			
		}
	} // class EnableTriggerCommand
	
	protected class DisableTriggerCommand extends Command {
		public void execute(Object params) {
			if (journalAction) {
				director.journal("Trigger action " + getAlias() 
						+ ", id= " + getID() 
						+ " was disabled");						
			}			
			// record that the action is disabled
			enabled= false;
			// process parameters
			processDisableTrigger(params);
		}
	} // class DisableTriggerCommand
	
	/** accessor for the enable trigger */
	public Command getEnableTriggerCommand() {
		return enableTriggerCommand;
	}
	
	/** accessor for the disable trigger */
	public Command getDisableTriggerCommand() {
		return disableTriggerCommand;
	}
	
	/**
	 * Is the action enabled?
	 * @return true if enabled, false if not
	 */
	public boolean isEnabled() {
		return enabled;
	}
	
} // TriggerAction
